using NyströmKME
using KernelFunctions
using DataFrames
using CSV
using FatDatasets
using Random
using StatsBase
using Base.Threads

function main_real_data(; nb_its=100)
	# ds_names = [("fma_mfcc", Dict())]
	# ds_names = [("fasttext_en", Dict(:ncols=>10^5))]
	# ds_names = [("intel_lab", Dict(:ncols=>10^5))]
	# ds_names = [("gowalla", Dict(:ncols=>10^5))]
	ds_names = [("intel_lab", Dict(:ncols=>10^5)),
				("gowalla", Dict(:ncols=>10^5)),
				("fasttext_en", Dict(:ncols=>10^5)),
				("fma_mfcc", Dict())]
	for (ds_name, kwargs) in ds_names
		X = if (ds_name == "gowalla") && (:ncols in keys(kwargs))
			Y = load_dataset(ds_name)
			@warn "Ignoring kwargs apart from ncols."
			Y[:,1:kwargs[:ncols]]
		else
			load_dataset(ds_name; kwargs...)
		end
		process_one_dataset(X, ds_name)
	end
end
function process_one_dataset(X, ds_name, n=10^4; nb_its=100, seed=1234)
	@info "Dataset $(ds_name), n=$n, sz=$(size(X,2))"
	mkpath(joinpath("results", ds_name))
	Random.seed!(seed)
	σ = estim_kernel_std(X[:,sample(1:size(X,2), 1000; replace=false)])
	k = gaussian_kernel(σ)

	# NyströmKME.set_max_memory!(200.0) # 150 Gb
	NyströmKME.set_max_memory!(6.0) 
	μ = EmpiricalKME(X, k)
	# NyströmKME.set_max_memory!(20.0) # for each thread…
	NyströmKME.set_max_memory!(0.5) # for each thread…
	
	base_output_name = joinpath("results", ds_name, "results_seed$(seed)_$(ds_name)")
	Threads.@threads for it = 1:nb_its
		output = base_output_name*"_it$(it).csv"
		m_ar = round.(Int, 10 .^ range(log10(2), min(3, log10(n)), length=8))
		m_already_computed = 0
		if isfile(output)
			output = output[1:end-4]*"_0.csv"
			while isfile(output)
				rr = match(r"(.*_)(.*).csv", output)
				num = parse(Int, rr[2]) + 1
				output = rr[1]*"$num.csv"
			end
			df = CSV.read(output, DataFrame)
			m_already_computed = maximum(df[.~ismissing.(df.m),:m])
		end
		if m_already_computed < maximum(m_ar)
			@debug "Iteration $it (thread $(Threads.threadid()))."
			df = DataFrame(exp=Int[], n=Int[], m=Union{Int, Missing}[], rkhs_error=Float64[], diff_rkhs_error=Float64[], method=String[]) 

			# Empirical estimator
			S = X[:,sample(1:size(X,2), n; replace=false)]
			μ_hat = EmpiricalKME(S, k)
			μhat_error = rkhs_dist(μ, μ_hat) 
			push!(df, Dict(:exp=>it, :n=>n, :m=>missing, :rkhs_error=>μhat_error, :diff_rkhs_error=>0.0,  :method=>"Empirical"))

			# Nyström estimator
			for m in m_ar
				if m > m_already_computed
					println("Iteration $it/$nb_its, n=$n, m=$m.")
					μ_Nys = Nyström_KME(S, k, m)
					μ_Nys_error = rkhs_dist(μ, μ_Nys) 
					derror = abs(μ_Nys_error - μhat_error)
					# store
					push!(df, Dict(:exp=>it, :n=>n, :m=>m, :rkhs_error=>μ_Nys_error, :diff_rkhs_error=>derror,  :method=>"Nyström"))
					CSV.write(output, df)
				else
					@info "Skipping m=$m for iteration $it, results already found."
				end
			end
			println("Saving result, df has size $(size(df))")
			CSV.write(output, df)
		else
			@info "Skipping iteration $it, results already found."
		end
	end
end
main_real_data()
