#######################################################################
#                               Metrics                               #
#######################################################################
#
function check_kernels(μ1, μ2)
	@assert μ1.k == μ2.k "The two mean embeddings do not belong to the same reproducing kernel Hilbert space."
end

"""
Computes [∫k(x,y₁)dμ(x), …, ∫k(x,y_n)dμ(x)] where Y=[y₁,…] for k(x,y) = exp(-‖x-y‖²/(2τ²)).
"""
function integral_kernel(p::DiagNormal, k::KF.TransformedKernel{KF.SqExponentialKernel{Distances.Euclidean},KF.ScaleTransform{Float64}}, Y)
	d = size(Y, 1)
	τ² = (1/k.transform.s[1])^2 						# k(x,y) = exp(-‖x-y‖²/(2τ²))
	q = MvNormal(mean(p), diagm(diag(p.Σ) .+ τ²))
	return (2*π*τ²)^(d/2)*pdf(q, Y) # Apply on each column of Y
end
function integral_kernel(p::MixtureModel, k, Y)
	[integral_kernel(p.components[i], k, Y) for i in 1:length(components(p))]' * probs(p)
end

"""
Computes ∫k(x,y)dp(x)dp(x) for k(x,y) = exp(-‖x-y‖²/(2τ²)) and p,q multivariate normal with diag covariances.
"""
function double_integral_kernel(p1::DiagNormal, p2::DiagNormal, k::KF.TransformedKernel{KF.SqExponentialKernel{Distances.Euclidean},KF.ScaleTransform{Float64}}, d)
	τ² = (1/k.transform.s[1])^2
	q = MvNormal(mean(p1), diagm(diag(p1.Σ)+diag(p2.Σ) .+ τ²))
	return (2π*τ²)^(d/2) * pdf(q, mean(p2))
end

"""
Computes ∫k(x,y)dp(x)dq(y) for k(x,y) = exp(-‖x-y‖²/(2τ²)) and p,q, gaussians.
"""
function double_integral_kernel(p1::MixtureModel, p2::MixtureModel, k::Kernel, d::Int)
	l1 = length(components(p1))
	l2 = length(components(p2))
	K = [double_integral_kernel(p1.components[i], p2.components[j], k, d) for i in 1:l1, j in 1:l2]
	return probs(p1)'*K*probs(p2)
end

function rkhs_ip(μ1::DistributionKME, μ2::DistributionKME)
	check_kernels(μ1, μ2)
	return double_integral_kernel(distribution(μ1), distribution(μ2),
									μ1.k, input_dimension(μ1))
end

function rkhs_ip(μ1::DistributionKME, μ2::DiscreteKME)
	check_kernels(μ1, μ2)
	X = data_samples(μ2)
	int_k = integral_kernel(distribution(μ1), μ1.k,  X)
	return int_k[:]' * weights(μ2)
end

"""
Returns the inner product between two discretely supported mean embeddings.
"""
function rkhs_ip(μ1::DiscreteKME, μ2::DiscreteKME)
	@debug "rkhs_ip between discrete distributions"
	check_kernels(μ1, μ2)
	rkhs_ip(data_samples(μ1), data_samples(μ2), weights(μ1), weights(μ2), μ1.k)
end
# For generic indexable types
rkhs_ip(X, Y, wX::Vector, wY::Vector, k::Kernel) = kernel_quadratic(k, X, Y, wX, wY)
function rkhs_ip(X::Matrix, Y::Matrix, wX::Vector, wY::Vector, k::Kernel)
	return if length(wX)*length(wY)*8/1024^3 > MAX_MEMORY_GB
		kernel_quadratic(k, X, Y, wX, wY)
	else
		wX'*kernelmatrix(k, X, Y; obsdim=2)*wY
	end
end
sq_rkhs_norm(μ::DiscreteKME) = sq_rkhs_norm(data_samples(μ), weights(μ), μ.k)
# For generic indexable types
sq_rkhs_norm(X, w::Vector, k::Kernel) = kernel_quadratic(k, X, w)
function sq_rkhs_norm(X::Matrix, w::Vector, k::Kernel)
	return if length(w)^2*8/1024^3 > MAX_MEMORY_GB
		kernel_quadratic(k, X, w)
	else
		w'*kernelmatrix(k, X; obsdim=2)*w
	end
end
# EXT Could be overloaded to optimize
sq_rkhs_norm(μ) = rkhs_ip(μ, μ)
sq_rkhs_dist(μ1, μ2) = abs(sq_rkhs_norm(μ1) + sq_rkhs_norm(μ2) - 2*rkhs_ip(μ1, μ2))
rkhs_dist(μ1, μ2) = sqrt(sq_rkhs_dist(μ1, μ2))
