abstract type KME end 

MAX_MEMORY_GB = 1.0
function set_max_memory!(x)
	global MAX_MEMORY_GB
	MAX_MEMORY_GB = x
end

"""
A Mean embedding stored via an associated discrete probability measure.
"""
abstract type DiscreteKME end 

struct NysKME{T} <: DiscreteKME
	Y::T
	k::KernelFunctions.Kernel
	w::Vector{Float64}
	d::Int
end

struct EmpiricalKME{T} <: DiscreteKME
	X::T
	k::KernelFunctions.Kernel
	d::Int
	sq_rkhs_norm::Float64
end

struct DistributionKME{T<:Distribution} <: KME
	p::T
	k::KernelFunctions.Kernel
	d::Int
end

#######################################################################
#                            Constructors                             #
#######################################################################

function EmpiricalKME(X, k)
	n = size(X,2)
	nrm = sq_rkhs_norm(X, ones(n)/n, k)
	EmpiricalKME(X, k, size(X,1), nrm)
end
sq_rkhs_norm(μ::EmpiricalKME) = μ.sq_rkhs_norm

function Nyström_KME(X, k, m) # EXT: add weighted sampling support for LS
	n = size(X,2)
	idx = sample(1:n, m; replace=false)
	Y = X[:, idx] 							# landmarks 
	K = kernelmatrix(k, Y; obsdim=2)
	# EXT One can use batching here if it doesn't fit in RAM
	Kmn = kernelmatrix(k, Y, X; obsdim=2) 
	w = pinv(K)* (Kmn*ones(n)/n)	
	# w = K \ (Kmn*ones(n)/n) 				# Weights 
	return NysKME(Y, k, w, size(X,1))
end

#######################################################################
#                              Utilities                              #
#######################################################################

support_size(μ::DiscreteKME) = size(data_samples(μ), 2)
data_samples(μ::EmpiricalKME) = μ.X
data_samples(μ::NysKME) = μ.Y
weights(μ::EmpiricalKME) = ones(support_size(μ))/support_size(μ)
weights(μ::NysKME) = μ.w

distribution(μ::DistributionKME) = μ.p
input_dimension(μ::KME) = μ.d

#######################################################################
#                            For iteration                            #
#######################################################################

"""
Structure for iterating jointly on X and Y.
"""
struct WeightedData{T1,T2}
	X::T1
	Y::Vector{T2}
end
function WeightedData(X,Y)
	@assert size(X,2) == length(Y)
	new(X, Y)
end
Base.size(D::WeightedData) = (size(D.X,1)+1, size(D.X,2))
Base.size(D::WeightedData, d) = size(D)[d]
Base.getindex(D::WeightedData, ::Colon, ::Colon) = getindex(D, :, 1:size(D,2))
Base.getindex(D::WeightedData, ::Colon, i2) = (D.X[:,i2], D.Y[i2])

function kernel_quadratic(k, X, Y, wX, wY; batchsize=nothing)
	bsz = (batchsize == nothing) ? choose_batchsize_quadratic(size(X,2), size(Y,2)) : batchsize
	# println("Using batch size $bsz")
	res = 0.0
	@debug "kernel_quadratic… "
	for (bX, bXw) in BatchIterator(WeightedData(X, wX); batchsize=bsz)
		for (bY, bYw) in BatchIterator(WeightedData(Y, wY); batchsize=bsz)
			res += bXw'*kernelmatrix(k, bX, bY)*bYw
		end
	end
	return res
end
function kernel_quadratic(k, X, wX; batchsize=nothing)
	bsz = (batchsize == nothing) ? choose_batchsize_quadratic(size(X,2), size(X,2)) : batchsize
	# println("Using batch size $bsz")
	res = 0.0
	@debug "kernel_quadratic (sym)… "
	for (i, (bX, bXw)) in enumerate(BatchIterator(WeightedData(X, wX); batchsize=bsz))
		for (j, (bY, bYw)) in enumerate(BatchIterator(WeightedData(X, wX); batchsize=bsz))
			if i==j
				res += bXw'*kernelmatrix(k, bX)*bXw
			elseif j>i
				res += 2*bXw'*kernelmatrix(k, bX, bY)*bYw
			end
		end
	end
	return res
end

# Assuming d<n, otherwise we might want to change that
function choose_batchsize_quadratic(nX,nY; maxbatchsize = 2^16, forcepow2 = true)
	fullsizeGB = nX*nY*sizeof(Float64)/1024^3	
	n = max(nX,nY)
	batchsize = (fullsizeGB > MAX_MEMORY_GB) ? ceil(Int, sqrt(MAX_MEMORY_GB*1024^3/sizeof(Float64))) : n
	batchsize = min(batchsize, maxbatchsize)
	@debug "Chosing batch size $(batchsize), MAX_MEMORY_GB=$(MAX_MEMORY_GB)"
	(forcepow2 && batchsize != n) ? prevpow(2, batchsize) : batchsize
end

